/*
 * command.c:
 *
 * Copyright (c) 2004 Chris Lightfoot. All rights reserved.
 * Email: chris@ex-parrot.com; WWW: http://www.ex-parrot.com/~chris/
 *
 */

static const char rcsid[] = "$Id: command.c,v 1.8 2004/09/29 15:52:05 chris Exp $";

/*
 * Commands: (preceded by ^A)
 *
 * ^A               send ^A
 *  h               help text
 *  v               report current settings
 *  b               send break
 *  x               disconnect
 *  n/o/e           set no/odd/even parity
 *  5/6/7/8         set number of data bits
 *  f n/h/s         set no/hardware/software flow control
 *  r nnnn ENTER    set baud rate
 */

#include <sys/types.h>

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>

#include "command.h"
#include "consolidate.h"
#include "tty.h"

static const char nottymsg[] =
    "\r\n"
    "    Not connected to a terminal -- no settings to report\r\n"
    "\r\n";

enum command_result show_tty_settings(const char *tty, int tty_fd, int client_fd) {
    if (!isatty(tty_fd)) {
        if (!write_to_client(tty, client_fd, nottymsg, (sizeof nottymsg) - 1))
            return r_client_error;
    } else {
        struct serial_options O;
        char buf[80];
        int n;
        if (!get_serial_options(tty_fd, &O))
            return r_tty_error;
        n = sprintf(buf, "\r\n    %s\r\n\r\n", describe_serial_options(&O));
        if (!write_to_client(tty, client_fd, buf, n))
            return r_client_error;
    }
    return r_finished;
}

int send_break(const char *tty, int fd) {
    struct termios T;
    speed_t old;
    if (!isatty(fd)) return 1;
    else if (-1 == tcgetattr(fd, &T)) return 0;
    old = cfgetispeed(&T);
    cfsetispeed(&T, B0);
    cfsetospeed(&T, B0);
    if (-1 == tcsetattr(fd, TCSANOW, &T))
        return 0;
    sleep(1);
    cfsetispeed(&T, old);
    cfsetospeed(&T, old);
    if (-1 == tcsetattr(fd, TCSANOW, &T))
        return 0;
    else {
        log_print(LOG_INFO, "%s: sent break", tty);
        return 1;
    }
}

int set_parity(const char *tty, int fd, const char c) {
    struct termios T;
    char *what;
    if (!isatty(fd)) return 1;
    else if (-1 == tcgetattr(fd, &T)) return 0;
    T.c_cflag &= ~(PARENB | PARODD);
    switch (c) {
        case 'n':
            what = "none";
            break; /* no change needed */
        case 'o':
            what = "odd";
            T.c_cflag |= PARODD | PARENB;
            break;
        case 'e':
            what = "even";
            T.c_cflag |= PARENB;
            break;
    }
    if (-1 == tcsetattr(fd, TCSANOW, &T))
        return 0;
    else {
        log_print(LOG_INFO, "%s: set parity to %s", tty, what);
        return 1;
    }
}

int set_data_bits(const char *tty, int fd, const int bits) {
    struct termios T;
    char *what;
    if (!isatty(fd)) return 1;
    else if (-1 == tcgetattr(fd, &T)) return 0;
    T.c_cflag &= ~CSIZE;
    switch (bits) {
        case 5:
            T.c_cflag |= CS5;
            break;
        case 6:
            T.c_cflag |= CS6;
            break;
        case 7:
            T.c_cflag |= CS7;
            break;
        case 8:
            T.c_cflag |= CS8;
            break;
        default:
            abort();    /* NOTREACHED */
    }
    if (-1 == tcsetattr(fd, TCSANOW, &T))
        return 0;
    else {
        log_print(LOG_INFO, "%s: set parity to %s", tty, what);
        return 1;
    }
}


int set_flowcontrol(const char *tty, int fd, const char c) {
    struct termios T;
    char *what;
    if (!isatty(fd)) return 1;
    else if (-1 == tcgetattr(fd, &T)) return 0;
    T.c_cflag &= ~(CRTSCTS);
    T.c_iflag &= ~(IXON | IXOFF);
    switch (c) {
        case 'n':
            what = "none";
            break;  /* no change needed */
        case 'h':
            what = "hardware";
            T.c_cflag |= CRTSCTS;
            break;
        case 's':
            what = "software";
            T.c_iflag |= IXON | IXOFF;
            break;
    }
    if (-1 == tcsetattr(fd, TCSANOW, &T))
        return 0;
    else {
        log_print(LOG_INFO, "%s: set flow control to %s", tty, what);
        return 1;
    }
}

int set_speed(const char *tty, int fd, const int speed) {
    struct termios T;
    speed_t s;
    if (!isatty(fd)) return 1;
    else if (-1 == tcgetattr(fd, &T)) return 0;
    s = n_to_rate(speed);
    cfsetispeed(&T, s);
    cfsetospeed(&T, s);
    if (-1 == tcsetattr(fd, TCSANOW, &T))
        return 0;
    else {
        log_print(LOG_INFO, "%s: set baud rate to %d", tty, speed);
        return 1;
    }
}

/* command_process TTY TTYFD CLIENT STATE C
 * Process the character C which has been received from the CLIENT socket,
 * returning a result and possibly modifying *STATE and the configuration
 * of the TTY through its file descriptor TTYFD. */
enum command_result command_process(const char *tty, int tty_fd, int client_fd, struct command_state *state, char c) {
#define leave(newstate, result)     do { state->s_state = newstate; return result; } while (0)
#define tty_error()                 do { leave(c_nocommand, r_tty_error); } while (0)
#define client_error()              do { leave(c_nocommand, r_client_error); } while (0)
#define bad_command()               do { leave(c_nocommand, r_badcommand); } while (0)
#define finished()                  do { leave(c_nocommand, r_finished); } while (0) /* XXX should send success message to CLIENT */
    extern const char helpmsg[];    /* in consolidate.c */
    switch (state->s_state) {
        case c_init:
            switch (tolower(c)) {
                case 1: /* ^A */
                    if (xwrite(tty_fd, "\x01", 1) != 1)
                        tty_error();
                    else
                        finished();

                case 'h':
                    if (!write_to_client(tty, client_fd, helpmsg, strlen(helpmsg)))
                        client_error();
                    else
                        finished();

                case 'v':
                    leave(c_nocommand, show_tty_settings(tty, tty_fd, client_fd));

                case 'b':
                    if (!send_break(tty, tty_fd))
                        tty_error();
                    else
                        finished();

                case 'x':
                    leave(c_init, r_disconnect);

                case 'n':
                case 'o':
                case 'e':
                    if (!set_parity(tty, tty_fd, tolower(c)))
                        tty_error();
                    else
                        finished();

                case '5':
                case '6':
                case '7':
                case '8':
                    if (!set_data_bits(tty, tty_fd, c - '0'))
                        tty_error();
                    else
                        finished();

                case 'f':
                    leave(c_flowcontrol, r_nothing);

                case 'r':
                    state->s_accum = 0;
                    leave(c_baud_digit, r_nothing);

                default:
                    bad_command();
            }

        case c_flowcontrol:
            switch (tolower(c)) {
                case 'n':
                case 'h':
                case 's':
                    if (!set_flowcontrol(tty, tty_fd, tolower(c)))
                        tty_error();
                    else
                        finished();

                default:
                    bad_command();
            }

        case c_baud_digit:
            if (c == '\n' || c == '\r') {
                if (B0 != n_to_rate(state->s_accum)) {
                    if (!set_speed(tty, tty_fd, state->s_accum))
                        tty_error();
                    else
                        finished();
                } else
                    bad_command();
            } else if (strchr("0123456789", c)) {
                state->s_accum *= 10;
                state->s_accum += c - '0';
                if (state->s_accum > 1000000)
                    bad_command();
                else
                    leave(c_baud_digit, r_nothing);
            } else
                bad_command();

        case c_nocommand:
        default:
            abort();    /* NOTREACHED */
    }
}
