/*
 * consolidate.c:
 * Console server thingy.
 *
 * Copyright (c) 2004 Chris Lightfoot. All rights reserved.
 * Email: chris@ex-parrot.com; WWW: http://www.ex-parrot.com/~chris/
 *
 */

static const char rcsid[] = "$Id: consolidate.c,v 1.27 2004/09/17 22:31:56 chris Exp $";

#include <sys/types.h>

#include <ctype.h>
#include <errno.h>
#include <pty.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <termios.h>
#include <string.h>
#include <unistd.h>

#include <sys/fcntl.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/un.h>
#include <sys/wait.h>

#include "command.h"
#include "consolidate.h"
#include "settings.h"
#include "tty.h"

/* 
 * Each process handles a given serial line. It reads from the serial line,
 * logging complete lines to a log file, and occasionally rotating that log
 * file when it becomes too large. The process also listens on a socket; users
 * connected to the socket receives a copy of data read from the serial line,
 * and any input they send is relayed to the serial line. Certain special
 * character sequences are interpreted as commands to change serial line
 * characteristics or whatever.
 */

int log_stderr = 1; /* true during startup */

#define make_non_blocking(fd)       fcntl((fd), F_SETFL, fcntl((fd), F_GETFL) | O_NONBLOCK)

/* open_tty TTY OPTIONS
 * Open a file descriptor on the named TTY, and configure it according to the
 * given OPTIONS. Returns a file descriptor on success, or -1 on failure. */
static int lock_tty = 1;
int open_tty(const char *tty, const struct serial_options *O) {
    int fd = -1;

    if (lock_tty && !uucp_lock(tty)) {
        log_print(LOG_ERR, "%s: unable to lock", tty);
        return -1;
    }

    if (-1 == (fd = open(tty, O_RDWR)))
        log_print(LOG_ERR, "%s: %m", tty);
    else if (!set_serial_options(fd, O))
        log_print(LOG_ERR, "%s: set_serial_options: %m", tty);
    else
        return fd;

    if (fd != -1)
        close(fd);
    return -1;
}

/* sigterm_handler N
 * Handle any signal which should make us terminate. */
static sig_atomic_t foad = 0;
static void sigterm_handler(int N) {
    foad = N;
}

/* sigchld_handler N
 * Handle SIGCHLD. This is easy as we only ever spawn one child. */
static sig_atomic_t child_terminated;
static int child_exit_status;
static pid_t child_pid;
static void sigchld_handler(int N) {
    pid_t P;
    int st;
    while ((P = waitpid(-1, &st, WNOHANG)) > 0 || (P == -1 && errno == EINTR)) {
        if (P == child_pid) {
            child_exit_status = st;
            child_terminated = 1;
        } else
            /* presumably a gzip child process or whatever */
            ;
    }
}

/* open_child COMMAND
 * Opens a child process, returning a file descriptor connected to its standard
 * input and output, or -1 on failure. */
static int run_in_pty = 1;
int open_child(const char *command) {
    int fd[2] = {-1, -1}, i;
    FILE *fp;

    if (!run_in_pty && -1 == socketpair(AF_UNIX, SOCK_STREAM, 0, fd)) 
        goto fail;

    xsignal(SIGCHLD, sigchld_handler);

    switch ((child_pid = run_in_pty ?
                    forkpty(&fd[0], NULL, NULL, NULL)
                  : fork())) {
        case 0:
            /* child */
            if (!run_in_pty) {
                close(0); close(1); close(2);
                dup(fd[1]); dup(fd[1]); dup(fd[1]);
            }
            for (i = getdtablesize(); i > 2; --i)
                close(i);
            execl("/bin/sh", "sh", "-c", command, NULL);

            /* In case of error, we want to report the error back to the parent. */
            fp = fdopen(2, "w");
            fprintf(fp, "consolidate: %s: %s\n", command, strerror(errno));
            _exit(1);

        case -1:
            goto fail;

        default:
            close(fd[1]);
            return fd[0];
    }

fail:
    close(fd[0]);
    close(fd[1]);
    return -1;
}

/* open_socket SOCKET
 * Return a file descriptor open and listening on the named SOCKET, or -1 on
 * failure. */
int open_socket(const char *s) {
    struct sockaddr_un sun = {0};
    int fd = -1;
#ifndef UNIX_PATH_MAX
#   define UNIX_PATH_MAX 108        /* seems even to be hardcoded in sys/un.h */
#endif
    if (strlen(s) > UNIX_PATH_MAX - 1) {
        log_print(LOG_ERR, "%s: UNIX socket names must be %d or fewer characters in length (sorry)", s, UNIX_PATH_MAX - 1);
        goto fail;
    }
    
    sun.sun_family = AF_UNIX;
    strcpy(sun.sun_path, s);

    if (-1 == (fd = socket(PF_UNIX, SOCK_STREAM, 0))) {
        log_print(LOG_ERR, "%s: socket: %m", s);
        goto fail;
    } else if (-1 == make_non_blocking(fd)) {
        log_print(LOG_ERR, "%s: O_NONBLOCK: %m", s);
        goto fail;
    }

    unlink(s);
    
    if (-1 == bind(fd, (struct sockaddr*)&sun, sizeof sun)) {
        log_print(LOG_ERR, "%s: bind: %m", s);
        goto fail;
    } else if (-1 == listen(fd, SOMAXCONN)) {
        log_print(LOG_ERR, "%s: listen: %m", s);
        goto fail;
    } else if (-1 == chmod(s, 0666)) {
        log_print(LOG_ERR, "%s: chmod(0666): %m", s);
        goto fail;
    }


    return fd;

fail:
    if (fd != -1) {
        close(fd);
        unlink(s);
    }
    return -1;
}

/* hrtime
 * Return high-precision time as a long double. */
long double hrtime(void) {
    struct timeval tv;
    long double D;
    gettimeofday(&tv, NULL);
    D = tv.tv_sec + 1e-6 * tv.tv_usec;
    return D;
}

const char helpmsg[] = /* not static as also used from command.c */
    "\r\n"
    "    consolidate: escape key sequences\r\n"
    "\r\n"
    "    ^A              send actual ^A to terminal\r\n"
    "     h              display this message\r\n"
    "     v              report current settings\r\n"
    "     b              send break\r\n"
    "     x              disconnect\r\n"
    "     n/o/e          set no/odd/even parity\r\n"
    "     5/6/7/8        set number of data bits\r\n"
    "     f n/h/s        set no/hardware/software flow control\r\n"
    "     r nnnn ENTER   set baud rate\r\n"
    "\r\n";

static const char replaymsg[] = 
    "    Press SPACE now to replay the last few bytes of output.\r\n"
    "    Press ANY OTHER KEY now to carry on without any replay.\r\n"
    "\r\n";

/* io_loop TAG TTY LOGNAME LOG SOCKETNAME SOCKET ACL
 * Actually run the console logging application. TTY, LOG and SOCKET are file
 * descriptors opened, respectively, on the terminal, log file, and
 * (optionally) the socket on which we listen for requests from users who wish
 * to interact with the terminal. If not NULL, ACL is a zero-terminated list of
 * UIDs who are permitted so to interact. Returns on error. TAG is a string
 * describing the TTY or command we're watching. */
static void io_loop(const char *tty, int serial_fd, const char *logfile, int logfile_fd, const char *sock, int sock_fd) {
    /*
     * Strategy: we just run round a select loop, processing input on each
     * descriptor; for the log file, we accumulate data in a buffer and flush
     * it to the log whenever it's full or we encounter a \n. New listeners on
     * the socket are stuffed in a linked list. We allow arbitrarily many, but
     * warn later users of the presence of others. We drop connections to any
     * clients if writes would block.
     */
    struct client {
        int c_fd;
        struct command_state c_state;
        struct client *c_prev, *c_next;
    } *clients = NULL;

    /* Log lines to a logfile, having accumulated them in a buffer. */
    char logbuf[2048];      /* we split lines longer than this */
    int o = 0;

    /* We offer to replay the last bit of output to a client when they connect.
     * This is done with a circular buffer. */
#define REPLAYLEN       1024
    struct {
        char r_buf[REPLAYLEN];
        int r_wrapped, r_off;
    } replay = {0};
    
    int select_errors = 0;
    int socket_errors = 0;

    time_t reopen_log_time = 0;

#define drop_client(C)  do { struct client *X; X = C; close(X->c_fd); X->c_fd = -1; log_print(LOG_INFO, "%s: disconnected client", tty); } while (0)
#define MAX(a, b)       ((a) > (b) ? (a) : (b))

    while (!foad) {
        struct timeval tv = {0};
        fd_set fds;
        int n = 1, N;
        struct client *C;

        tv.tv_sec = 10;

        FD_ZERO(&fds);
        
        FD_SET(serial_fd, &fds);
        n = MAX(serial_fd + 1, n);

        if (sock_fd != -1) {
            FD_SET(sock_fd, &fds);
            n = MAX(sock_fd + 1, n);
            for (C = clients; C; C = C->c_next)
                if (C->c_fd != -1) {
                    FD_SET(C->c_fd, &fds);
                    n = MAX(C->c_fd + 1, n);
                }
        }

        if (-1 == (N = select(n, &fds, NULL, NULL, &tv) && errno != EINTR && errno != EAGAIN)) {
            log_print(select_errors >= 10 ? LOG_ERR : LOG_WARNING, "%s: select: %m", tty);
            if (select_errors++ >= 10) {
                log_print(LOG_ERR, "%s: too many errors from select; aborting", tty);
                goto quit;
            }
        } else
            select_errors = 0;

        /* Report termination of child. */
        if (child_pid != 0 && child_terminated) {
            if (WIFEXITED(child_exit_status))
                log_print(LOG_ERR, "%s: child PID %d exited with status %d; aborting", tty, (int)child_pid, WEXITSTATUS(child_exit_status));
            else
                log_print(LOG_ERR, "%s: child PID %d killed by signal %d; aborting", tty, (int)child_pid, WTERMSIG(child_exit_status));
            goto quit;
        }

        if (N == 0)
            continue;

        /* Data from console to log file and clients. */
        if (FD_ISSET(serial_fd, &fds)) {
            ssize_t n;
            n = xread(serial_fd, logbuf + o, sizeof(logbuf) - o);
            if (n == -1) {
                log_print(LOG_ERR, "%s: read: %m; aborting", tty);
                goto quit;
            } else if (n > 0) {
                char *p;
                int j;

                /* Copy data from terminal to connected clients. */
                for (C = clients; C; C = C->c_next)
                    /* Don't write to clients in the `not started' state. */
                    if (C->c_fd != -1 && C->c_state.s_state != c_notstarted && !write_to_client(tty, C->c_fd, logbuf + o, n))
                        drop_client(C);

                /* Update circular buffer. */
                for (j = o; j < o + n; ++j) {
                    replay.r_buf[replay.r_off] = logbuf[j];
                    replay.r_off = (replay.r_off + 1) % REPLAYLEN;
                    if (replay.r_off == 0)
                        replay.r_wrapped = 1;
                }

                o += n;

                /* Consider whether we've got a whole line of text from the
                 * console. */
                if ((p = memchr(logbuf, '\n', o)))
                    ++p; /* advance past LF */
                else if (o == sizeof(logbuf))
                    p = logbuf + o;

                if (p) {
                    size_t a;
                    a = p - logbuf;
                    if (a > 0) {
                        /* have some text to write */

                        /* If we ever fail to write to the logfile, close it
                         * and defer reopening it for ten seconds so we don't
                         * spam the syslog with boring messages about free disk
                         * space or whatever. */
                        if (logfile_fd == -1
                            && time(NULL) >= reopen_log_time
                            && -1 == (logfile_fd = open(logfile, O_WRONLY | O_APPEND | O_SYNC | O_CREAT, 0666))) {
                            log_print(LOG_ERR, "%s: %s: %m", tty, logfile);
                            reopen_log_time = time(NULL) + 10;
                        }

                        if (logfile_fd != -1 && -1 == (logfile_fd = write_to_logfile(tty, logfile, logfile_fd, logbuf, a)))
                            reopen_log_time = time(NULL) + 10;
                    }

                    memmove(logbuf, p, sizeof(logbuf) - a);
                    o -= a;
                }
            }
        }

        /* New connections on socket. */
        if (sock_fd != -1 && FD_ISSET(sock_fd, &fds)) {
            int f;
            if (-1 != (f = accept(sock_fd, NULL, NULL))) {
                struct client *cnew;
                socket_errors = 0;
                
                alloc_struct(client, cnew);
                cnew->c_fd = f;
                cnew->c_state.s_state = c_notstarted;
                cnew->c_prev = NULL;
                cnew->c_next = clients;
                if (clients) clients->c_prev = cnew;
                clients = cnew;
                if (!write_to_client(tty, cnew->c_fd, helpmsg, strlen(helpmsg))
                    || !write_to_client(tty, cnew->c_fd, replaymsg, strlen(replaymsg)))
                    drop_client(cnew);
                log_print(LOG_INFO, "%s: new client connected", tty);
            } else if (errno != EINTR && errno != EAGAIN) {
                ++socket_errors;
                log_print(LOG_ERR, "%s: socket %s: accept: %m", tty, sock);

                if (socket_errors >= 10) {
                    log_print(LOG_CRIT, "%s: %s: too many socket errors; closing socket (no new clients will be able to connect)", tty, sock);
                    close(sock_fd);
                    sock_fd = -1;
                }
            }
        }
        
        /* Drop disconnected clients. */
        for (C = clients; C;) {
            if (C->c_fd == -1) {
                struct client *C2;
                if (C->c_prev)
                    C->c_prev->c_next = C->c_next;
                else
                    clients = C->c_next;
                if ((C2 = C->c_next))
                    C->c_next->c_prev = C->c_prev;
                xfree(C);
                C = C2;
            } else
                C = C->c_next;
        }

        /* Read from clients. */
        for (C = clients; C; C = C->c_next) {
            char ibuf[256];
            ssize_t m;
            size_t i;

            if (!FD_ISSET(C->c_fd, &fds))
                continue;

            if ((m = xread(C->c_fd, ibuf, sizeof ibuf)) <= 0) {
                if (m == -1)
                    log_print(LOG_ERR, "%s: socket %s: connected client: %m", tty, sock);
                else
                    log_print(LOG_INFO, "%s: socket %s: client disconnected", tty, sock);
                close(C->c_fd);
                C->c_fd = -1;
                continue;
            }

            /* Now have m bytes of input from the client. Process this
             * input byte-by-byte, interpreting escape sequences as
             * appropriate. */
            for (i = 0; C->c_fd != -1 && i < m; ++i) {
                if (C->c_state.s_state == c_notstarted) {
                    /* If the client's first keypress is SPACE, we replay the
                     * last bit of output. Otherwise we don't. */
                    if (ibuf[i] == ' ') {
                        if (replay.r_wrapped
                            && !write_to_client(tty, C->c_fd, replay.r_buf + replay.r_off, REPLAYLEN - replay.r_off))
                            drop_client(C);
                        if (replay.r_off > 0
                            && !write_to_client(tty, C->c_fd, replay.r_buf, replay.r_off))
                            drop_client(C);
                    }
                    C->c_state.s_state = c_nocommand;
                } else if (C->c_state.s_state == c_nocommand) {
                    /* Process input. */
                    if (ibuf[i] == 1)
                        C->c_state.s_state = c_init;
                    else if (!write_to_tty(serial_fd, ibuf + i, 1)) {
                        log_print(LOG_ERR, "%s: write: %m; aborting", tty);
                        goto quit;
                    }
                } else {
                    /* Process commands. */
                    enum command_result R;
                    struct serial_options O;
                    switch ((R = command_process(tty, serial_fd, C->c_fd, &C->c_state, ibuf[i]))) {
                        case r_nothing:
                            break;
                            
                        case r_finished:
                            /* There might have been a change of setting; save it. */
                            if (isatty(serial_fd) && get_serial_options(serial_fd, &O))
                                save_default_settings(tty, &O);
                            break;

                        case r_badcommand:
                            if (write_to_client(tty, C->c_fd, "\a", 1))
                                break;
                            /* fall through on error */

                        case r_client_error:
                            log_print(LOG_ERR, "%s: connected client: %m", tty);
                            drop_client(C);
                            break;

                        case r_tty_error:
                            log_print(LOG_ERR, "%s: client command: %m; aborting", tty);
                            goto quit;

                        case r_disconnect:
                            log_print(LOG_ERR, "%s: client disconnected on command", tty);
                            drop_client(C);
                            break;
                    }
                }
            }
        }
    }
quit:
    /* Write any last gasps to the log file. */
    if (o > 0 && logfile_fd != -1)
        logfile_fd = write_to_logfile(tty, logfile, logfile_fd, logbuf, o);
}

/* is_a_terminal STRING
 * Does STRING refer to the name of a terminal? */
static int is_a_terminal(const char *s) {
    int fd, f;
    if (-1 == (fd = open(s, O_RDWR)))
        return 0;
    f = isatty(fd);
    close(fd);
    return f;
}

/* usage STREAM
 * Print a usage message to STREAM. */
void usage(FILE *fp) {
    fprintf(fp,
"consolidate: console server daemon\n"
"\n"
"Usage: consolidate -h | [OPTIONS] (TTY | COMMAND) LOGFILE [SOCKET]\n"
"\n"
"TTY gives the name of the serial device to which consolidate will attach;\n"
"alternatively, if a COMMAND is given, consolidate runs it inside a\n"
"pseudoterminal rather than connecting to a real terminal. LOGFILE names a\n"
"file in which lines received from the TTY or COMMAND are logged. Optionally,\n"
"SOCKET gives the path of a UNIX socket to which client processes may connect\n"
"to interact with the TTY or COMMAND.\n"
"\n"
"OPTIONS are:\n"
"\n"
" -h             Display this help message.\n"
" -s SETTING     Setting for TTY, in the form rate,parity,data-bits; the\n"
"                default is whatever is saved in ~/.consolidate, or otherwise\n"
"                \"9600,n,8\". Ignored if a COMMAND is given.\n"
" -S SETTING     Like -s, but don't override saved settings.\n"
" -f FLOW        Flow control setting; either \"none\", \"hw\" for RTS/CTS,\n"
"                or \"sw\" for XON/XOFF; the default is whatever is saved in\n"
"                ~/.consolidate, or otherwise \"hw\". Ignored if a COMMAND is\n"
"                given.\n"
" -F FLOW        Like -f, but don't override saved settings.\n"
" -P             If executing a COMMAND, don't run it inside a pseudoterminal\n"
"                but just connect to its standard input/output/error. Ignored\n"
"                if a TTY is given.\n"
" -u             Don't attempt to acquire the UUCP lock file for a TTY.\n"
" -l SIZE        The largest size, in bytes, to which LOGFILE may grow before\n"
"                it is rotated and replaced with a new one.\n"
" -d             Do not detach from controlling terminal.\n"
" -e             Log errors to standard error, as well as the system log.\n"
"\n"
"Consolidate, version " CONSOLIDATE_VERSION ".\n"
"Copyright (c) 2004 Chris Lightfoot.\n"
"%s\n",
        rcsid);
}

/* main ARGC ARGV
 * Entry point. */
int main(int argc, char *argv[]) {
    static const char optstr[] = "+hs:S:f:F:Pul:dec:";
    char *ser_opts = NULL, *fc_opt = NULL;
    int ser_opts_override = 1, fc_opt_override = 1;
    struct serial_options lineopts = {0};
    int c, n;
    extern size_t maxlogsize; /* in io.c */
    int detach = 1, do_log_stderr = 0;
    char *tty, *logfile, *sock, *command = NULL;
    int serial_fd, logfile_fd, sock_fd = -1;
    
    maxlogsize = 1024 * 1024;

    opterr = 0;

    xsignal(SIGPIPE, SIG_IGN);
    xsignal(SIGTERM, sigterm_handler);
    xsignal(SIGINT, sigterm_handler);

    while (-1 != (c = getopt(argc, argv, optstr))) {
        switch (c) {
            case 'h':
                usage(stdout);
                return 0;

            case 'S':
                ser_opts_override = 0;
                /* fall through */
            case 's':
                ser_opts = optarg;
                break;

            case 'F':
                fc_opt_override = 0;
                /* fall through */
            case 'f':
                fc_opt = optarg;
                break;

            case 'P':
                run_in_pty = 0;
                break;

            case 'u':
                lock_tty = 0;
                break;

            case 'l':
                if ((n = atoi(optarg)) < 1024) {
                    log_print(LOG_ERR, "\"%s\": bad maximum logfile size; try -h for help", optarg);
                    return 1;
                } else {
                    maxlogsize = (size_t)n;
                    break;
                }

            case 'd':
                detach = 0;
                break;

            case 'e':
                do_log_stderr = 1;
                break;

            case '?':
            default:
                if (strchr(optstr, optopt))
                    log_print(LOG_ERR, "option -%c requires an argument; try -h for help", optopt);
                else
                    log_print(LOG_ERR, "unknown option -%c; try -h for help", optopt);
                return 1;
        }
    }

    if (!(tty = argv[optind++])) {
        log_print(LOG_ERR, "you must specify a serial line or a command; try -h for help");
        return 1;
    }

    if (!is_a_terminal(tty)) {
        command = tty;
        tty = NULL;
    } else {
        int saved_settings;
        saved_settings = load_default_settings(tty, &lineopts);

        /* Maybe override saved settings? */
        if (ser_opts && (!saved_settings || ser_opts_override) && !parse_serial_options(ser_opts, &lineopts)) {
            log_print(LOG_ERR, "\"%s\": invalid serial line settings; try -h for help", ser_opts);
            return 1;
        }

        if (fc_opt && (!saved_settings || fc_opt_override)) {
            if (0 == strcasecmp("hw", fc_opt))
                lineopts.so_flowcontrol = f_hw;
            else if (0 == strcasecmp("sw", fc_opt))
                lineopts.so_flowcontrol = f_sw;
            else if (0 == strcasecmp("none", fc_opt))
                lineopts.so_flowcontrol = f_none;
            else {
                log_print(LOG_ERR, "\"%s\": invalid flow-control setting; try -h for help", fc_opt);
                return 1;
            }
        }
    }

    if (!(logfile = argv[optind++])) {
        log_print(LOG_ERR, "you must specify a log file; try -h for help");
        return 1;
    } else if (-1 == (logfile_fd = open(logfile, O_WRONLY | O_APPEND | O_SYNC | O_CREAT, 0666))) {
        log_print(LOG_ERR, "%s: %m", logfile);
        return 1;
    } else if (!command && -1 == (serial_fd = open_tty(tty, &lineopts))) {
        log_print(LOG_ERR, "%s: %m", tty);
        return 1;
    } else if (command && -1 == (serial_fd = open_child(command))) {
        log_print(LOG_ERR, "%s: %m", command);
        return 1;
    }
 
    if ((sock = argv[optind++])) {
        if (-1 == (sock_fd = open_socket(sock))) {
            log_print(LOG_ERR, "%s: %m", sock);
            log_print(LOG_ERR, "will proceed with logging only (no tty interaction)");
        } else {
            if (argv[optind++])
                log_print(LOG_WARNING, "ignored extra command-line parameters");
        }
    }

    /* Initialisation over; should log messages to stderr only if wanted. */
    log_stderr = do_log_stderr;

    io_loop(tty ? tty : command, serial_fd, logfile, logfile_fd, sock, sock_fd); /* only returns on error */

    if (tty) {
        if (lock_tty)
            uucp_unlock(tty);
    } else if (!child_terminated)
        /* Kill the child, just in case. */
        kill(child_pid, SIGKILL);

    return 1;
}
