/*
 * settings.c:
 * Save/load serial port settings.
 *
 * Copyright (c) 2004 Chris Lightfoot. All rights reserved.
 * Email: chris@ex-parrot.com; WWW: http://www.ex-parrot.com/~chris/
 *
 */

static const char rcsid[] = "$Id: settings.c,v 1.4 2004/08/14 16:49:20 chris Exp $";

#include <sys/param.h>

#include <ctype.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "consolidate.h"
#include "tty.h"

/* get_conffile_name
 * Return the name of consolidate's configuration file. */
static char *get_conffile_name(void) {
    static char buf[PATH_MAX];
    char *s;

    if ((s = getenv("CONSOLIDATE_DEFAULTS")))
        return s;

    if (!(s = getenv("HOME"))) {
        struct passwd *p;
        if ((p = getpwuid(getuid())))
            s = p->pw_dir;
        else
            return NULL;
    }
    
    sprintf(buf, "%s/.consolidate", s);
    return buf;
}

static const struct serial_options defaults = {
    B9600, 8, p_none, f_hw
};

#define WHITESPACE " \t"

/* load_default_settings TTY SETTINGS
 * Obtain the default SETTINGS for TTY, either by loading them from a
 * ~/.consolidate file, or simply returning the defaults. Returns true if
 * the settings were read from a file, otherwise false. */
int load_default_settings(const char *tty, struct serial_options *O) {
    char *conffile;
    int ret = 0;

    *O = defaults;
    if ((conffile = get_conffile_name())) {
        FILE *fp;
        char *s;
        if (!(fp = fopen(conffile, "rt")))
            return 0;

        while ((s = get_line(fp))) {
            char *hash, *thistty, *settings, *fc;
            if ((hash = strchr(s, '#')))
                *hash = 0;
            
            thistty = s + strspn(s, WHITESPACE);
            if (!*thistty) continue;
            
            settings = thistty + strcspn(thistty, WHITESPACE);
            *settings++ = 0;

            if (0 != strcmp(thistty, tty))
                continue;
            
            settings += strspn(settings, WHITESPACE);
            if (!*settings) continue;

            fc = settings + strcspn(settings, WHITESPACE);
            *fc++ = 0;

            fc += strspn(fc, WHITESPACE);
            if (!*fc) continue;

            if (!parse_serial_options(settings, O)) {
                *O = defaults;
                break;
            }

            switch (tolower(*fc)) {
                case 'n':
                    O->so_flowcontrol = f_none;
                    break;

                case 'h':
                    O->so_flowcontrol = f_hw;
                    break;

                case 's':
                    O->so_flowcontrol = f_sw;
                    break;

                default:
                    *O = defaults;
                    break;
            }
            ret = 1;
            break;
        }

        fclose(fp);
    }

    return ret;
}

static const char conffileblurb[] =
"#\n"
"# consolidate(8) settings file\n"
"#\n"
"# This file records the default settings consolidate should use for tty\n"
"# devices. \"#\" introduces a comment; otherwise, each line should consist\n"
"# of the name of a device, followed by settings for it in the form required\n"
"# by consolidate's -s and -f options; for instance,\n"
"#\n"
"#   /dev/ttyS0 115200,n,8 hw\n"
"#\n"
"# sets the first serial port to 115200 bps, no parity, 8 data bits, hardware\n"
"# flow control.\n"
"#\n"
"# consolidate will update lines in this file when you change settings from\n"
"# within the program.\n"
"#\n"
"# By default, consolidate stores settings in ~/.consolidate, but you can\n"
"# override this by setting the environment variable CONSOLIDATE_DEFAULTS to\n"
"# specify some other filename.\n"
"#\n";

/* save_default_settings TTY SETTINGS
 * Save SETTINGS as the defaults for TTY. */
void save_default_settings(const char *tty, const struct serial_options *O) {
    char *conffile, *newconffile = NULL;
    int fd = -1;
    FILE *fp_in = NULL, *fp_out = NULL;
    int savedit = 0;

    if (!(conffile = get_conffile_name()))
        return;

    newconffile = sprintf_new("%s-XXXXXX", conffile);

    if (-1 == (fd = mkstemp(newconffile)))
        goto fail;
    else if (!(fp_out = fdopen(fd, "wt")))
        goto fail;

    if ((fp_in = fopen(conffile, "rt"))) {
        char *s;
        size_t n;
        n = strlen(tty);
        while ((s = get_line(fp_in))) {
            char *hash, *thistty;

            if ((hash = strchr(s, '#')))
                *hash = 0;

            thistty = s + strspn(s, WHITESPACE);
            if (strcspn(thistty, WHITESPACE) == n && 0 == strncmp(thistty, tty, n)) {
                *thistty = 0;
                fprintf(fp_out, "%s%s %d,%c,%d %s\n",
                        s,  /* replicate indentation */
                        tty,
                        rate_to_n(O->so_rate),
                        O->so_parity == p_none ? 'n' :
                            (O->so_parity == p_odd ? 'o' : 'e'),
                        O->so_databits,
                        O->so_flowcontrol == f_none ? "none" :
                            (O->so_flowcontrol == f_hw ? "hw" : "sw"));
                savedit = 1;
            } else {
                if (hash) *hash = '#';
                fprintf(fp_out, "%s\n", s);
            }
        }

        if (ferror(fp_in) || ferror(fp_out)) goto fail;
    } else
        fwrite(conffileblurb, (sizeof conffileblurb) - 1, 1, fp_out);

    if (!savedit)
        fprintf(fp_out, "%s %d,%c,%d %s\n",
                tty,
                rate_to_n(O->so_rate),
                O->so_parity == p_none ? 'n' :
                    (O->so_parity == p_odd ? 'o' : 'e'),
                O->so_databits,
                O->so_flowcontrol == f_none ? "none" :
                    (O->so_flowcontrol == f_hw ? "hw" : "sw"));

    if (ferror(fp_out))
        goto fail;

    fclose(fp_out);
    fp_out = NULL;
    if (0 == rename(newconffile, conffile))
        return;

fail:
    if (newconffile) {
        unlink(newconffile);
        xfree(newconffile);
    }
    if (fp_in) fclose(fp_in);
    if (fp_out) fclose(fp_out);
    else if (-1 != fd) close(fd);
}
