/*
 * util.c:
 * Useful functions.
 *
 * Copyright (c) 2003 Chris Lightfoot. All rights reserved.
 * Email: chris@ex-parrot.com; WWW: http://www.ex-parrot.com/~chris/
 *
 */

static const char rcsid[] = "$Id: util.c,v 1.5 2004/06/22 19:22:47 chris Exp $";

#include <sys/types.h>

#include <errno.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "consolidate.h"

/* xmalloc:
 * Malloc, and abort if malloc fails. */
void *xmalloc(size_t n) {
    void *v;
    v = malloc(n);
    if (!v) abort();
    return v;
}

/* xcalloc:
 * As above. */
void *xcalloc(size_t n, size_t m) {
    void *v;
    v = calloc(n, m);
    if (!v) abort();
    return v;
}

/* xrealloc:
 * As above. */
void *xrealloc(void *w, size_t n) {
    void *v;
    v = realloc(w, n);
    if (n != 0 && !v) abort();
    return v;
}

/* xstrdup:
 * As above. */
char *xstrdup(const char *s) {
    char *t;
    t = strdup(s);
    if (!t) abort();
    return t;
}

/* xfree:
 * Free, ignoring a passed NULL value. */
void xfree(void *v) {
    if (v) free(v);
}

/* xsignal NUMBER HANDLER
 * Set a signal, with a similar interface to signal(2) using sigaction(2). */
void (*xsignal(int signum, void(*handler)(int)))(int) {
    struct sigaction sa = {0}, sa_old;
    sa.sa_handler = handler;
    sa.sa_flags = SA_RESTART;
    if (sigaction(signum, &sa, &sa_old) == -1)
        return SIG_ERR;
    else
        return sa_old.sa_handler;
}

/* sprintf_new FORMAT [ARG ...]
 * Like sprintf, but return a newly allocated string of sufficient size to hold
 * the formatted output. */
char *sprintf_new(const char *fmt, ...) {
    /* Cf. snprintf(3). */
    size_t size = 256;
    char *p;
    va_list ap;
    p = xmalloc(size);
    while (1) {
        int n;
        /* Try to print in the allocated space. */
        va_start(ap, fmt);
        n = vsnprintf (p, size, fmt, ap);
        va_end(ap);
        /* If that worked, return the string. */
        if (n > -1 && (size_t)n < size)
            return p;
        /* Else try again with more space. */
        if (n > -1)     /* glibc 2.1 */
            size = n + 1;   /* precisely what is needed */
        else            /* glibc 2.0 */
            size *= 2;      /* twice the old size */
        p = xrealloc(p, size);
    }
}

/* get_line STREAM
 * Read a line of text from STREAM, returning NULL on error or end-of-file, or
 * a string on success. Any carriage return is stripped. */
char *get_line(FILE *fp) {
    static char *buf;
    static size_t buflen;
    int c;
    size_t i;
    if (!buf)
        buf = xmalloc(buflen = 256);

    i = 0;
    while ((c = fgetc(fp)) != EOF) {
        if (c == '\n')
            break;
        else if (i == buflen - 1)    /* space for terminating 0 */
            buf = xrealloc(buf, buflen *= 2);
        buf[i++] = (char)c;
    }
    buf[i] = 0;
    if (c == EOF && (i == 0 || ferror(fp)))
        return NULL;
    else
        return buf;
}
